"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.lambdaWorker = void 0;

var _runtimeBase = require("@resolve-js/runtime-base");

var _schedulerFactory = require("./scheduler-factory");

var _eventSubscriberNotifierFactory = require("./event-subscriber-notifier-factory");

var _metrics = require("./metrics");

var _websocketEventHandler = require("./websocket-event-handler");

var _apiGatewayHandler = require("./api-gateway-handler");

var _cloudServiceEventHandler = require("./cloud-service-event-handler");

var _schedulerEventHandler = require("./scheduler-event-handler");

var _lodash = _interopRequireDefault(require("lodash.partial"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const log = (0, _runtimeBase.getLog)('lambda-worker');
const GRACEFUL_WORKER_SHUTDOWN_TIME = 30 * 1000;

const getLambdaVacantTime = lambdaContext => lambdaContext.getRemainingTimeInMillis() - GRACEFUL_WORKER_SHUTDOWN_TIME;

const WORKER_HTTP_REQUEST_DURATION = 25 * 1000;
let coldStart = true;

const createCloudRuntime = async (coldStartContext, lambdaContext, subscriberInterface, overrideFactoryParams = {}) => {
  const getVacantTimeInMillis = (0, _lodash.default)(getLambdaVacantTime, lambdaContext, () => void 0);
  const {
    eventstoreAdapter: eventStoreAdapterFactory,
    readModelConnectors: readModelConnectorsFactories
  } = coldStartContext.assemblies;
  const runtimeParams = { ...coldStartContext,
    eventStoreAdapterFactory,
    readModelConnectorsFactories,
    getVacantTimeInMillis,
    notifyEventSubscriber: subscriberInterface.notifyEventSubscriber,
    invokeBuildAsync: subscriberInterface.invokeBuildAsync,
    ensureQueue: subscriberInterface.ensureQueue,
    deleteQueue: subscriberInterface.deleteQueue,
    getEventSubscriberDestination: subscriberInterface.getEventSubscriberDestination,
    ...overrideFactoryParams
  };
  const runtime = await (0, _runtimeBase.createRuntime)(runtimeParams);
  const scheduler = await (0, _schedulerFactory.schedulerFactory)(runtime, coldStartContext.domainInterop.sagaDomain.schedulerName); // TODO: only remaining late binding, protected by guard

  log.debug(`scheduler late binding`);
  runtimeParams.scheduler = scheduler;
  return {
    runtime,
    scheduler,
    getVacantTimeInMillis
  };
};

const lambdaWorker = async (coldStartContext, lambdaEvent, lambdaContext) => {
  log.debug('executing application lambda');
  log.verbose(JSON.stringify(lambdaEvent, null, 2));
  lambdaContext.callbackWaitsForEmptyEventLoop = false;
  const lambdaRemainingTimeStart = lambdaContext.getRemainingTimeInMillis();
  const subscriberInterface = await (0, _eventSubscriberNotifierFactory.eventSubscriberNotifierFactory)({
    eventSubscriberScope: coldStartContext.eventSubscriberScope,
    lambdaContext
  });
  const makeRuntime = (0, _lodash.default)(createCloudRuntime, coldStartContext, lambdaContext, subscriberInterface);
  let runtime = null; // TODO: too complex

  try {
    if (lambdaEvent.resolveSource === 'DeployService') {
      log.debug('identified event source: deployment service');
      void ({
        runtime
      } = await makeRuntime({}));
      const executorResult = await (0, _cloudServiceEventHandler.handleCloudServiceEvent)(lambdaEvent, runtime, {
        domain: coldStartContext.domain,
        domainInterop: coldStartContext.domainInterop,
        eventListeners: coldStartContext.eventListeners,
        eventSubscriberScope: coldStartContext.eventSubscriberScope,
        deleteQueue: subscriberInterface.deleteQueue,
        ensureQueue: subscriberInterface.ensureQueue,
        getEventSubscriberDestination: subscriberInterface.getEventSubscriberDestination,
        upstream: true,
        performanceTracer: coldStartContext.performanceTracer
      });
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);
      return executorResult;
    } else if (lambdaEvent.resolveSource === 'BuildEventSubscriber') {
      var _coldStartContext$eve;

      log.debug('identified event source: event-subscriber-direct');
      void ({
        runtime
      } = await makeRuntime({}));
      const {
        resolveSource,
        eventSubscriber,
        ...buildParameters
      } = lambdaEvent;
      void resolveSource;
      const currentEventSubscriber = (_coldStartContext$eve = coldStartContext.eventListeners.get(eventSubscriber)) !== null && _coldStartContext$eve !== void 0 ? _coldStartContext$eve : {
        isSaga: false
      };
      let executorResult = await runtime.eventSubscriber.build({ ...buildParameters,
        eventSubscriber,
        coldStart
      });
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);

      if (currentEventSubscriber.isSaga) {
        await (0, _eventSubscriberNotifierFactory.waitForSubscriber)(true);
        executorResult = await runtime.eventSubscriber.build({ ...buildParameters,
          eventSubscriber,
          coldStart
        });
        log.verbose(`sagaExecutorResult: ${JSON.stringify(executorResult)}`);
      }

      return executorResult;
    } else if (Array.isArray(lambdaEvent.Records) && [...new Set(lambdaEvent.Records.map(record => record != null ? record.eventSource : null))].every(key => key === 'aws:sqs')) {
      const errors = [];
      const records = lambdaEvent.Records.map(record => {
        try {
          return JSON.parse(record.body);
        } catch (err) {
          errors.push(new Error(`Invalid record ${JSON.stringify(record)} parsing failed with error ${err}`));
          return null;
        }
      });
      void ({
        runtime
      } = await makeRuntime({}));
      let buildParameters = {
        coldStart,
        eventsWithCursors: []
      };

      for (const record of records) {
        if (record == null || record.eventSubscriber == null || record.eventSubscriber.constructor !== String || !(record.event == null && record.cursor == null || record.event != null && record.event.constructor === Object && record.cursor != null && record.cursor.constructor === String)) {
          errors.push(new Error(`Malformed record ${JSON.stringify(record)}`));
          continue;
        }

        const {
          eventSubscriber,
          event,
          cursor,
          ...notification
        } = record;

        if (buildParameters.eventSubscriber == null) {
          buildParameters.eventSubscriber = eventSubscriber;
        } else if (buildParameters.eventSubscriber !== eventSubscriber) {
          errors.push(new Error(`Multiple event subscribers ${buildParameters.eventSubscriber} and ${eventSubscriber} are not allowed in one window`));
          continue;
        }

        if (event != null && cursor != null) {
          void buildParameters.eventsWithCursors.push({
            event,
            cursor
          });
        }

        Object.assign(buildParameters, notification);
      }

      if (buildParameters.eventsWithCursors.length === 0) {
        buildParameters.eventsWithCursors = null;
      }

      if (errors.length > 0) {
        const summaryError = new Error(errors.map(({
          message
        }) => message).join('\n'));
        summaryError.stack = errors.map(({
          stack
        }) => stack).join('\n');
        throw summaryError;
      }

      const executorResult = await runtime.eventSubscriber.build(buildParameters);
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);
      return executorResult;
    } else if (lambdaEvent.resolveSource === 'Scheduler') {
      log.debug('identified event source: cloud scheduler');
      const data = await makeRuntime({});
      const executorResult = await (0, _schedulerEventHandler.handleSchedulerEvent)(lambdaEvent, data.scheduler);
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);
      return executorResult;
    } else if (lambdaEvent.resolveSource === 'Websocket') {
      log.debug('identified event source: websocket');
      void ({
        runtime
      } = await makeRuntime({}));
      const executorResult = await (0, _websocketEventHandler.handleWebsocketEvent)(lambdaEvent, runtime);
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);
      return executorResult;
    } else if (lambdaEvent.headers != null && lambdaEvent.httpMethod != null) {
      const overrideParams = lambdaEvent.requestStartTime !== undefined && Number.isSafeInteger(lambdaEvent.requestStartTime) ? {
        getVacantTimeInMillis: () => {
          var _lambdaEvent$requestS;

          return ((_lambdaEvent$requestS = lambdaEvent.requestStartTime) !== null && _lambdaEvent$requestS !== void 0 ? _lambdaEvent$requestS : 0) + WORKER_HTTP_REQUEST_DURATION - Date.now();
        }
      } : {};
      const data = await makeRuntime(overrideParams);
      runtime = data.runtime;
      runtime.eventStoreAdapter.establishTimeLimit(data.getVacantTimeInMillis);
      log.debug('identified event source: API gateway');
      log.verbose(JSON.stringify(lambdaEvent.httpMethod, null, 2), JSON.stringify(lambdaEvent.headers, null, 2));
      const executorResult = await (0, _apiGatewayHandler.handleApiGatewayEvent)(lambdaEvent, lambdaContext, runtime, {
        performanceTracer: coldStartContext.performanceTracer,
        buildTimeConstants: coldStartContext.constants,
        routesTrie: coldStartContext.routesTrie,
        domain: coldStartContext.domain,
        domainInterop: coldStartContext.domainInterop,
        eventSubscriberScope: coldStartContext.eventSubscriberScope,
        seedClientEnvs: coldStartContext.seedClientEnvs,
        eventListeners: coldStartContext.eventListeners
      });
      log.verbose(`executorResult: ${JSON.stringify(executorResult)}`);
      return executorResult;
    } else {
      void ({
        runtime
      } = await makeRuntime({}));
      throw new Error(`abnormal lambda execution on event ${JSON.stringify(lambdaEvent)}`);
    }
  } catch (error) {
    var _runtime, _runtime$monitoring;

    log.error(`top-level event handler execution error! Event: ${JSON.stringify(lambdaEvent)}`);
    (_runtime = runtime) === null || _runtime === void 0 ? void 0 : (_runtime$monitoring = _runtime.monitoring) === null || _runtime$monitoring === void 0 ? void 0 : _runtime$monitoring.group({
      Part: 'Internal'
    }).error(error);

    if (error instanceof Error) {
      log.error('error', error.message);
      log.error('error', error.stack);
    } else {
      log.error(JSON.stringify(error));
    }

    throw error;
  } finally {
    if (runtime != null) {
      await runtime.dispose();
    }

    if (process.env.RESOLVE_PERFORMANCE_MONITORING) {
      await (0, _metrics.putDurationMetrics)(lambdaEvent, lambdaContext, coldStart, lambdaRemainingTimeStart);
    }

    coldStart = false;
    log.debug('reSolve framework was disposed');
  }
};

exports.lambdaWorker = lambdaWorker;
//# sourceMappingURL=lambda-worker.js.map